/*----------------------------------------------------------------------
  File    : ptlset.c
  Contents: prototype-less fuzzy clustering
  Author  : Christian Borgelt
  History : 2007.01.08 file created
            2007.01.10 first version of ptl_update() completed
            2007.01.14 function ptl_desc() added
            2007.01.16 function ptl_parse() added
            2007.01.17 first full version completed
            2007.01.23 stored number of objects to consider
            2007.02.27 offset for membership degrees added
            2013.07.18 adapted to definitions ATTID, VALID, TPLID etc.
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "ptlset.h"
#ifdef STORAGE
#include "storage.h"
#endif

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define BLKSIZE   16            /* block size for reallocation */
#define EPSILON   1e-12         /* minimum degree of membership */

/* --- error codes --- */
#define E_MSDCNT  (-16)         /* wrong number of membership degrees */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
static const char *msgs[] = {   /* error messages */
  /*      0 to  -7 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /*     -8 to -15 */  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
  /* E_MSDCNT  -16 */  "#wrong number of membership degrees",
};

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static double** matrix (DIMID rowcnt, DIMID colcnt)
{                               /* --- create a matrix */
  DIMID  k;                     /* loop variable */
  double **mat;                 /* created matrix */
  double *p;                    /* to organize the memory */

  mat = (double**)malloc((size_t)rowcnt *sizeof(double*));
  if (!mat) {            return NULL; }
  p = (double*)malloc((size_t)rowcnt *(size_t)colcnt *sizeof(double));
  if (!p)   { free(mat); return NULL; }
  for (k = 0; k < rowcnt; k++){ /* allocate a rowcnt x colcnt matrix */
    mat[k] = p; p += colcnt; }  /* and organize the memory */
  return mat;                   /* return the created matrix */
}  /* matrix() */

/*--------------------------------------------------------------------*/

static int powers (PTLSET *ptl)
{                               /* --- compute powers of m.s. degrees */
  DIMID  i, k;                  /* loop variables */
  double *m, *p;                /* to traverse the matrix rows */

  if (!ptl->pows) {             /* if there is no such buffer yet */
    ptl->pows = matrix(ptl->rowcnt, ptl->clscnt);
    if (!ptl->pows) return -1;  /* create a buffer to store the */
  }                             /* powers of the membership degrees */
  for (k = 0; k < ptl->rowcnt; k++) {
    m = ptl->msds[k];           /* traverse the matrix rows for */
    p = ptl->pows[k];           /* membership degrees and powers */
    i = ptl->clscnt;            /* get the number of clusters */
    if      (ptl->msexp == 1) { /* if to use m.s. degrees directly */
      while (--i >= 0) p[i] = m[i]                  +ptl->msoff; }
    else if (ptl->msexp == 2) { /* if to square the m.s. degrees */
      while (--i >= 0) p[i] = m[i] *m[i]            +ptl->msoff; }
    else {                      /* if other membership exponent */
      while (--i >= 0) p[i] = pow(m[i], ptl->msexp) +ptl->msoff;
    }                           /* compute the needed powers */
  }                             /* of the membership degrees */
  return 0;                     /* return 'ok' */
}  /* powers() */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

PTLSET* ptl_create (DIMID rowcnt, DIMID clscnt)
{                               /* --- create a p.t.-less cluster set */
  PTLSET *ptl;                  /* created prototype-less cluster set */

  assert((clscnt > 0) && (clscnt <= rowcnt));
  ptl = (PTLSET*)malloc(sizeof(PTLSET));
  if (!ptl) return NULL;        /* create the base structure */
  ptl->clscnt = clscnt;         /* store the number of clusters */
  ptl->rowcnt = rowcnt;         /* and the number of tuples */
  ptl->objcnt = 0;
  ptl->norm   = 0;
  ptl->msexp  = 2;              /* set the default exponents */
  ptl->msoff  = 0;              /* and the default offset */
  ptl->tfexp  = 1;
  ptl->objval = NAN;
  ptl->wgts   = NULL;           /* create the membership matrix */
  ptl->msds   = matrix(rowcnt+1, clscnt);
  ptl->buf    = (ptl->msds) ? ptl->msds[0] : NULL;
  ptl->pows   = NULL;           /* create buffers for execution */
  ptl->dists  = (double*)malloc((size_t)rowcnt *sizeof(double));
  ptl->ids    = (DIMID*) malloc((size_t)rowcnt *sizeof(DIMID));
  if (!ptl->msds || !ptl->dists || !ptl->ids) {
    ptl_delete(ptl); return NULL; }
  return ptl;                   /* return the created cluster set */
}  /* ptl_create() */

/*--------------------------------------------------------------------*/

void ptl_delete (PTLSET *ptl)
{                               /* --- delete a p.t.-less cluster set */
  if (ptl->pows) {              /* delete the powers buffer */
    free(ptl->pows[0]); free(ptl->pows); }
  if (ptl->ids)   free(ptl->ids);   /* delete the buffers */
  if (ptl->dists) free(ptl->dists); /* needed for execution */
  if (ptl->wgts)  free(ptl->wgts);  /* delete the weight vector */
  if (ptl->buf)   free(ptl->buf);   /* delete the membership matrix */
  if (ptl->msds)  free(ptl->msds);  /* and its vector of rows */
  free(ptl);                    /* delete the base structure */
}  /* ptl_delete() */

/*--------------------------------------------------------------------*/

void ptl_msexp (PTLSET *ptl, double msexp, double msoff)
{                               /* --- set the membership exponent */
  assert(ptl);                  /* check the function arguments */
  if ((msexp == ptl->msexp) && (msoff == ptl->msoff)) return;
  if (!isnan(msexp) && (msexp >  0))
    ptl->msexp = msexp;         /* store the new membership exponent */
  if (!isnan(msoff) && (msoff > -1))
    ptl->msoff = msoff;         /* store the new membership offset */
  if (ptl->pows) {              /* delete the buffer for powers */
    free(ptl->pows[0]); free(ptl->pows); ptl->pows = NULL; }
}  /* ptl_msexp() */

/*--------------------------------------------------------------------*/

void ptl_tfexp (PTLSET *ptl, double tfexp)
{                               /* --- set the transform. exponent */
  assert(ptl);                  /* check the function arguments */
  if (!isnan(tfexp) && (tfexp >= 0))
    ptl->tfexp = tfexp;         /* store the new transform. exponent */
}  /* ptl_tfexp() */

/*--------------------------------------------------------------------*/

int ptl_setwgt (PTLSET *ptl, DIMID rowid, double wgt)
{                               /* --- set weight of a data object */
  DIMID i;                      /* loop variable */

  assert(ptl                    /* check the function arguments */
  &&    (rowid >= 0) && (rowid < ptl->rowcnt));
  if (!ptl->wgts) {             /* if there is no weight vector */
    if (wgt == 1) return 0;     /* check whether a vector is needed */
    ptl->wgts = (double*)malloc((size_t)ptl->rowcnt *sizeof(double));
    if (!ptl->wgts) return -1;  /* create a vector of object weights */
    for (i = 0; i < ptl->rowcnt; i++)
      ptl->wgts[i] = 1;         /* init. the created weight vector */
  }                             /* with unit weights (default) */
  ptl->wgts[rowid] = wgt;       /* set the given weight */
  return 0;                     /* return 'ok' */
}  /* ptl_setwgt() */

/*--------------------------------------------------------------------*/

void ptl_init (PTLSET *ptl, int mode, RANDFN randfn)
{                               /* --- initialize m.s. degrees */
  DIMID  i, k, n, t;            /* loop variables, buffers */
  double *p, d;                 /* to traverse the membership degrees */
  double sum;                   /* sum of the membership degrees */
  DIMID  *ids;                  /* vector of indices */

  assert(ptl);                  /* check the function arguments */
  if (mode & PTL_UNIFORM) {     /* if to set uniform m.s. degrees */
    d = 1.0/(double)ptl->clscnt;/* compute uniform membership degree */
    for (k = 0; k < ptl->rowcnt; k++) {
      p = ptl->msds[k];         /* traverse the matrix rows */
      for (i = 0; i < ptl->clscnt; i++)
        p[i] = d;               /* set all membership degrees */
    } }                         /* to the same value */
  else {                        /* if to set random m.s. degrees */
    assert(randfn);             /* check the function arguments */
    for (k = 0; k < ptl->rowcnt; k++) {
      p = ptl->msds[k];         /* traverse the matrix rows */
      for (sum = 0, i = 0; i < ptl->clscnt; i++)
        sum += p[i] = randfn(); /* init. and sum membership degrees */
      if ((sum <= 0) || (sum == 1))
        continue;               /* check whether normalization needed */
      sum = 1.0/sum;            /* compute normalization factor */
      for (i = 0; i < ptl->clscnt; i++)
        p[i] *= sum;            /* normalize the membership degrees */
    }                           /* to sum 1 (satisfy constraint) */
  }
  if ((mode & PTL_SEED) == 0)   /* if not to select seed objects, */
    return;                     /* abort the function */
  assert(randfn);               /* check the function arguments */
  ids = ptl->ids;               /* get the internal buffer */
  for (k = n = ptl->rowcnt; --k >= 0; )
    ids[k] = k;                 /* fill buffer with object indices */
  for (k = 0; k < ptl->clscnt; k++) {
    i = (DIMID)((double)n *randfn());
    if (i > --n) i = n;         /* compute a random index and */
    if (i < 0)   i = 0;         /* exchange selected index with last */
    t = ids[i]; ids[i] = ids[n]; ids[n] = t;
    p = ptl->msds[t];           /* get the corresponding matrix row */
    memset(p, 0, (size_t)ptl->clscnt *sizeof(double));
    p[k] = 1;                   /* assign the row to cluster k */
  }                             /* (choose one object per cluster) */
}  /* ptl_init() */

/*--------------------------------------------------------------------*/

double ptl_update (PTLSET *ptl, DISTMAT *mat,
                   DIMID *order, RANDFN randfn)
{                               /* --- update a p.t.-less cluster set */
  DIMID  i, j, k, n;            /* loop variables */
  double **pows;                /* powers of membership degrees */
  double *m, *p;                /* to traverse the matrix rows */
  double d, sum;                /* buffers for distances / sums */
  double max = 0;               /* maximal change of a m.s. degree */

  assert(ptl && mat);           /* check the function arguments */
  if (!ptl->pows                /* if the powers are not computed */
  &&  (powers(ptl) != 0))       /* and the buffer cannot be created, */
    return -1;                  /* abort the function */

  /* --- build index permutation --- */
  if (!order && randfn) {       /* if to shuffle the order */
    order = ptl->ids;           /* get the internal buffer */
    for (k = 0; k < ptl->rowcnt; k++)
      order[k] = k;             /* fill buffer with object indices */
    for (k = ptl->rowcnt; k > 0; ) {
      i = (DIMID)((double)k *randfn());
      if (i > --k) i = k;       /* compute a random index and */
      if (i < 0)   i = 0;       /* make sure it is in [0,n-1] */
      j = order[i]; order[i] = order[k]; order[k] = j;
    }                           /* exchange selected index with last */
  }                             /* (create random index order) */

  /* --- update membership degrees --- */
  pows = ptl->pows;             /* get the precomputed powers */
  m = ptl->msds[ptl->rowcnt];   /* traverse the matrix rows */
  for (n = 0; n < ptl->rowcnt; n++) {
    j = (order) ? order[n] : n; /* traverse the data objects */
    for (i = 0; i < ptl->clscnt; i++)
      m[i] = 0;                 /* clear the membership degrees */
    if (mat->objcnt > 0) {      /* if restricted number of objects */
      for (k = mat->objcnt; --k >= 0; ) {
        d = mat->dists[j][k];   /* traverse the selected objects */
        p = pows[i = mat->ids[j][k]];
        if (ptl->wgts) d *= ptl->wgts[i];
        for (i = 0; i < ptl->clscnt; i++)
          m[i] += p[i] *d;      /* sum the distances weighted with */
      } }                       /* the cluster membership degrees */
    else {                      /* if unrestricted number of objects */
      for (k = 0; k < ptl->rowcnt; k++) {
        if (k == j) continue;   /* traverse all other matrix rows */
        d = (k > j) ? mat->dists[k][j] : mat->dists[j][k];
        if (ptl->wgts) d *= ptl->wgts[k];
        p = pows[k];            /* get the data point distance */
        for (i = 0; i < ptl->clscnt; i++)
          m[i] += p[i] *d;      /* sum the distances weighted with */
      }                         /* the cluster membership degrees */
    }
    if (ptl->tfexp <= 0) {      /* if hard assignment to clusters */
      k = 0; p = pows[j];       /* initialize the best cluster index */
      for (i = 0; i < ptl->clscnt; i++) {
        if (m[i] < m[k]) k = i; /* find the best cluster and */
        p[i] = m[i] = 0;        /* clear all membership degrees, */
      }                         /* then set the membership degree */
      p[k] = m[k] = 1;          /* of the best cluster to 1 */
      p = ptl->msds[j];         /* determine whether an assignment */
      if (p[k] < 1) max = 1;    /* of a data object changed */
      ptl->msds[j] = m; m = p;  /* swap updated vector into place */
      continue;                 /* continue with the next object */
    }
    for (i = 0; i < ptl->clscnt; i++)
      m[i] = (m[i] >= EPSILON) ? 1.0/m[i] : 1.0/EPSILON;
    sum = 0;                    /* compute reciprocal values */
    i = ptl->clscnt;            /* get the number of clusters */
    if      (ptl->tfexp == 1) { /* if unit transformation exponent */
      while (--i >= 0) sum += m[i]; }
    else if (ptl->tfexp == 2) { /* if quadratic transformation */
      while (--i >= 0) sum += m[i] *= m[i]; }
    else {                      /* if other transformation exponent */
      while (--i >= 0) sum += m[i] = pow(m[i], ptl->tfexp);
    }                           /* sum the adapted update terms */
    if ((sum > 0) && (sum != 1)) {
      sum = 1.0/sum;            /* compute the normalization factor */
      for (i = 0; i < ptl->clscnt; i++)
        m[i] *= sum;            /* traverse the membership degrees */
    }                           /* and normalize them to sum 1 */
    for (i = 0; i < ptl->clscnt; i++)
      if (m[i] < EPSILON) m[i] = 0;
    p = pows[j];                /* get the corresponding powers */
    i = ptl->clscnt;            /* and the number of clusters */
    if      (ptl->msexp == 1) { /* if to use m.s. degrees directly */
      while (--i >= 0) p[i] = m[i]                  +ptl->msoff; }
    else if (ptl->msexp == 2) { /* if to square the m.s. degrees */
      while (--i >= 0) p[i] = m[i] *m[i]            +ptl->msoff; }
    else {                      /* if other membership exponent */
      while (--i >= 0) p[i] = pow(m[i], ptl->msexp) +ptl->msoff;
    }                           /* (re)compute the needed powers */
    p = ptl->msds[j];           /* compare to old membership degrees */
    for (i = 0; i < ptl->clscnt; i++) {
      d = fabs(m[i] -p[i]);     /* traverse all membership degrees */
      if (d > max) max = d;     /* compute the maximum difference */
    }                           /* between a new value and the old */
    ptl->msds[j] = m; m = p;    /* swap updated vector into place */
  }                             /* (avoid copying the vector) */
  ptl->msds[ptl->rowcnt] = m;   /* store last row as new buffer */
  return max;                   /* return maximal m.s. degree change */
}  /* ptl_update() */

/*--------------------------------------------------------------------*/

void ptl_regular (PTLSET *ptl)
{                               /* --- regularize cluster weight */
  DIMID  i, k;                  /* loop variables */
  double t;                     /* buffer for computation */
  double *msds;                 /* to traverse the membership degrees */
  double *sums;                 /* vector of membership sums */

  sums = ptl->msds[ptl->rowcnt];/* get and clear the summation buffer */
  memset(sums, 0, (size_t)ptl->clscnt *sizeof(double));
  for (k = 0; k < ptl->rowcnt; k++) {
    msds = ptl->msds[k];        /* traverse the matrix rows */
    for (i = 0; i < ptl->clscnt; k++)
      sums[i] += msds[i];       /* sum the membership degrees */
  }                             /* for all classes */
  t = (double)ptl->rowcnt /(double)ptl->clscnt;
  for (i = 0; i < ptl->clscnt; i++)
    sums[i] = t /sums[i];       /* compute regularization factors */
  for (k = 0; k < ptl->rowcnt; k++) {
    msds = ptl->msds[k];        /* traverse the matrix rows */
    for (t = 0, i = 0; i < ptl->clscnt; i++)
      t += msds[i] *= sums[i];  /* reg. and sum membership degrees */
    if ((t <= 0) || (t == 1))   /* check whether a normalization of */
      continue;                 /* the membership degrees is needed */
    t = 1.0/t;                  /* compute the normalization factor */
    for (i = 0; i < ptl->clscnt; i++)
      msds[i] *= t;             /* normalize the membership degrees */
  }                             /* to sum 1 (satisfy constraint) */
}  /* ptl_regular() */

/*--------------------------------------------------------------------*/

double ptl_objfn (PTLSET *ptl, DISTMAT *mat)
{                               /* --- compute objective function */
  DIMID  i, j, k;               /* loop variables */
  double *pj, *pk;              /* powers of membership degrees */
  double d;                     /* buffer for distances */
  double sum = 0;               /* sum of terms, objective function */

  if (!ptl->pows                /* if the powers are not computed */
  &&  (powers(ptl) != 0))       /* and the buffer cannot be created, */
    return -1;                  /* abort the function */
  for (j = ptl->rowcnt; --j > 0; ) {
    pj = ptl->pows[j];          /* traverse all data objects */
    if (mat->objcnt > 0) {      /* if restricted number of objects */
      for (k = mat->objcnt; --k >= 0; ) {
        pk = ptl->pows[i = mat->ids[j][k]];
        d  = mat->dists[j][k];  /* get distances of objects */
        if (ptl->wgts) d *= ptl->wgts[i];
        for (i = 0; i < ptl->clscnt; i++)
          sum += pj[i] *pk[i] *d;     /* sum the terms of */
      } }                       /* the objective function */
    else {                      /* if unrestricted number of objects */
      for (k = 0; k < j; k++) { /* traverse all data object pairs */
        pk = ptl->pows[k];      /* get the powers of the m.s. degrees */
        d  = mat->dists[j][k];  /* and the distance of the objects */
        if (ptl->wgts) d *= ptl->wgts[k];
        for (i = 0; i < ptl->clscnt; i++)
          sum += pj[i] *pk[i] *d;
      }                         /* sum the terms of */
    }                           /* the objective function */
  }
  return ptl->objval = sum;     /* return the objective function */
}  /* ptl_objfn() */

/*--------------------------------------------------------------------*/

DIMID ptl_exec (PTLSET *ptl, double *dists,
                DIMID *ids, DIMID objcnt, double *msds)
{                               /* --- execute a cluster set */
  DIMID  i, k, n;               /* loop variables, buffer */
  double d;                     /* to traverse the distances */
  double *m;                    /* to traverse the membership degrees */
  double sum, max;              /* sum/maximum of membership degrees */

  assert(ptl && dists           /* check the function arguments */
  &&    ((objcnt == 0) || (abs(objcnt) >= ptl->rowcnt) || ids));
  if (objcnt < 0) objcnt = -objcnt;
  if (!msds)                    /* get internal buffer if necessary */
    msds = ptl->msds[ptl->rowcnt];  /* and init. membership degrees */
  memset(msds, 0, (size_t)ptl->clscnt *sizeof(double));
  n = ((objcnt == 0) || (objcnt > ptl->rowcnt))
    ? ptl->rowcnt : objcnt;     /* get the number of objects */
  for (k = 0; k < n; k++) {     /* traverse the distances */
    d = dists[k];               /* to the different data objects */
    if (d <= 0) continue;       /* skip vanishing distances */
    m = ptl->msds[i = (objcnt == 0) ? k : ids[k]];
    if (ptl->wgts) d *= ptl->wgts[i];
    i = ptl->clscnt;            /* get corresp. membership degrees */
    if      (ptl->msexp == 1) { /* if to use m.s. terms directly */
      while (--i >= 0) msds[i] += d *(m[i]                +ptl->msoff);}
    else if (ptl->msexp == 2) { /* if to square the membership terms */
      while (--i >= 0) msds[i] += d *(m[i] *m[i]          +ptl->msoff);}
    else {                      /* if other exponent of m.s. degree */
      while (--i >= 0) msds[i] += d *(pow(m[i],ptl->msexp)+ptl->msoff);
    }                           /* sum the membership terms */
  }                             /* for the different classes */
  max = 0; k = 0; m = msds;     /* initialize the best cluster index */
  for (i = 0; i < ptl->clscnt; i++) {
    m[i] = (m[i] >= EPSILON) ? 1.0/m[i] : 1.0/EPSILON;
    if (m[i] > max) { max = m[i]; k = i; }
  }                             /* find index of best cluster */
  if (m == ptl->msds[ptl->rowcnt])
    return k;                   /* check whether normalization needed */
  if (ptl->tfexp <= 0) {        /* if hard assignment to clusters */
    memset(m, 0, (size_t)ptl->clscnt *sizeof(double));
    m[k] = 1; return k;         /* set the membership degree */
  }                             /* of the best cluster to 1 */
  sum = 0;                      /* init. sum for normalization */
  i = ptl->clscnt;              /* get the number of clusters */
  if      (ptl->tfexp == 1) {   /* if unit normalization exponent */
    while (--i >= 0) sum += m[i]; }
  else if (ptl->tfexp == 2) {   /* if unit normalization exponent */
    while (--i >= 0) sum += m[i] *= m[i]; }
  else {                        /* if other membership exponent */
    while (--i >= 0) sum += m[i] = pow(m[i], ptl->tfexp);
  }                             /* sum the adapted values */
  if ((sum > 0) && (sum != 1)){ /* if a normalization is needed, */
    sum = 1.0/sum;              /* compute the normalization factor */
    for (i = 0; i < ptl->clscnt; i++)
      m[i] *= sum;              /* normalize the membership degrees */
  }
  for (i = 0; i < ptl->clscnt; i++)
    if (m[i] < EPSILON) m[i] = 0;
  return k;                     /* return index of best cluster */
}  /* ptl_exec() */

/*--------------------------------------------------------------------*/

int ptl_desc (PTLSET *ptl, FILE *file, int mode, int maxlen)
{                               /* --- describle a cluster set */
  DIMID  i, k;                  /* loop variables */
  int    len, l;                /* loop variables for comments */
  double *m;                    /* to traverse the membership degrees */

  assert(ptl && file);          /* check the function arguments */

  /* --- print a header (as a comment) --- */
  if (mode & PTL_TITLE) {       /* if the title flag is set */
    len = (maxlen > 0) ? maxlen -2 : 70;
    fputs("/*", file); for (l = len; --l >= 0; ) fputc('-', file);
    fputs("\n  prototype-less cluster set\n", file);
    for (l = len; --l >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* print a title header */
  }                             /* (as a comment) */

  /* --- print the cluster set --- */
  fputs("ptlset = {\n", file);  /* start the description */
  fputs("  params = ", file);   /* write the parameters */
  fprintf(file, "%.16g, %.16g,", ptl->msexp, ptl->msoff);
  fprintf(file, " %.16g, %"DIMID_FMT, ptl->tfexp, ptl->objcnt);
  if (ptl->norm) fputs(", normalized", file);
  fputs(";\n", file);           /* write flag for normalized table */
  fputs("  msdegs = {\n", file);
  for (k = 0; k < ptl->rowcnt; k++) {
    m = ptl->msds[k];           /* traverse the matrix row */
    if (k > 0) fputs(" },\n", file);
    fputs("    { ", file);      /* enclose each row in curly braces */
    for (i = 0; i < ptl->clscnt; i++) {
      if (i > 0) fputs(", ", file);
      #if 0
      fprintf(file, "%.16g", m[i]);
      #else
      fprintf(file, "%g", m[i]);
      #endif
    }                           /* traverse the matrix row and */
  }                             /* print the membership degrees */
  fputs(" } };\n", file);       /* terminate the membership matrix */
  fputs("};\n",   file);        /* terminate the description */

  /* --- print additional information (as a comment) --- */
  if (mode & PTL_INFO) {        /* if the add. info. flag is set */
    len = (maxlen > 0) ? maxlen -2 : 70;
    fputs("\n/*", file); for (l = len; --l >= 0; ) fputc('-', file);
    fprintf(file, "\n  number of clusters: %"DIMID_FMT, ptl->clscnt);
    fprintf(file, "\n  number of tuples  : %"DIMID_FMT, ptl->rowcnt);
    fprintf(file, "\n  number of objects : %"DIMID_FMT, ptl->objcnt);
    fputc('\n', file);          /* print additional cluster info. */
    if (!isnan(ptl->objval))    /* and objective function value */
      fprintf(file, "  objective function: %g\n", ptl->objval);
    for (l = len; --l >= 0; ) fputc('-', file);
    fputs("*/\n", file);        /* (as a comment) */
  }
  return ferror(file);          /* return the write status */
}  /* ptl_desc() */

/*--------------------------------------------------------------------*/

static int parse (PTLSET *ptl, SCANNER *scan)
{                               /* --- parse a p.t.-less cluster set */
  DIMID  i;                     /* loop variable */
  DIMID  cnt, vsz;              /* number of m.s. degs., vector size */
  double d, *m = NULL;          /* to traverse the m.s. degrees */

  assert(ptl && scan);          /* check the function arguments */
  if ((scn_token(scan) != T_ID) /* check for 'ptlset' */
  ||  (strcmp(scn_value(scan), "ptlset") != 0))
    SCN_ERROR(scan, E_STREXP, "ptlset");
  SCN_NEXT(scan);               /* consume 'ptlset' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */
  if ((scn_token(scan) != T_ID) /* check for 'params' */
  ||  (strcmp(scn_value(scan), "params") != 0))
    SCN_ERROR(scan, E_STREXP, "params");
  SCN_NEXT(scan);               /* consume 'params' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
  ptl->msexp = strtod(scn_value(scan), NULL);
  if (ptl->msexp <= 0)          SCN_ERROR(scan, E_NUMBER);
  SCN_NEXT(scan);               /* get the membership exponent */
  SCN_CHAR(scan, ',');          /* consume ',' */
  if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
  ptl->msoff = strtod(scn_value(scan), NULL);
  SCN_NEXT(scan);               /* get the membership offset */
  SCN_CHAR(scan, ',');          /* consume ',' */
  if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
  ptl->tfexp = strtod(scn_value(scan), NULL);
  if (ptl->tfexp <  0)          SCN_ERROR(scan, E_NUMBER);
  SCN_NEXT(scan);               /* get the transformation exponent */
  SCN_CHAR(scan, ',');          /* consume ',' */
  if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
  ptl->objcnt = strtodim(scn_value(scan), NULL);
  if (ptl->objcnt < 0) ptl->objcnt = -ptl->objcnt;
  SCN_NEXT(scan);               /* get the number of objects */
  if (scn_token(scan) == ',') { /* if another parameter follows */
    SCN_NEXT(scan);             /* consume ',' */
    if ((scn_token(scan) != T_ID)  /* check for 'normalized' */
    ||  (strcmp(scn_value(scan), "normalized") != 0))
      SCN_ERROR(scan, E_STREXP, "normalized");
    SCN_NEXT(scan);             /* consume 'normalized' */
    ptl->norm = 1;              /* set the normalization flag */
  }
  SCN_CHAR(scan, ';');          /* consume ';' */
  if ((scn_token(scan) != T_ID) /* check for 'msdegs' */
  ||  (strcmp(scn_value(scan), "msdegs") != 0))
    SCN_ERROR(scan, E_STREXP, "msdegs");
  SCN_NEXT(scan);               /* consume 'msdegs' */
  SCN_CHAR(scan, '=');          /* consume '=' */
  SCN_CHAR(scan, '{');          /* consume '{' */
  ptl->clscnt = cnt = vsz = 0;  /* init. the number of clusters */
  while (1) {                   /* membership read loop */
    SCN_CHAR(scan, '{');        /* consume '{' */
    i = 0;                      /* init. the membership counter */
    while (1) {                 /* read a row of the matrix */
      if (scn_token(scan) != T_NUM) SCN_ERROR(scan, E_NUMEXP);
      d = strtod(scn_value(scan), NULL);
      if ((d < 0) || (d > 1))       SCN_ERROR(scan, E_NUMBER);
      SCN_NEXT(scan);           /* get and consume membership degree */
      if (cnt >= vsz) {         /* if the vector is full */
        vsz += (vsz > BLKSIZE) ? vsz >> 1 : BLKSIZE;
        m = (double*)realloc(ptl->buf,
              (size_t)(vsz+ptl->clscnt) *sizeof(double));
        if (!m) SCN_ERROR(scan, E_NOMEM);/* compute new vector size, */
        ptl->buf = m;           /* then allocate an enlarged vector */
      }                         /* and set the new vector */
      ptl->buf[cnt++] = d;      /* store the membership degree */
      i++;                      /* and count them */
      if (scn_token(scan) != ',') break;
      SCN_NEXT(scan);           /* if at end of m.s. list, abort */
    }                           /* the loop, otherwise consume ',' */
    SCN_CHAR(scan, '}');        /* consume '}' */
    if      (ptl->clscnt <= 0) ptl->clscnt = cnt;
    else if (ptl->clscnt != i) SCN_ERROR(scan, E_MSDCNT);
    if (scn_token(scan) != ',') break;
    SCN_NEXT(scan);             /* if at end of row list, abort */
  }                             /* the loop, otherwise consume ',' */
  ptl->rowcnt = cnt / ptl->clscnt;
  if (cnt < vsz) {              /* try to shrink the m.s. vector */
    cnt += ptl->clscnt;         /* to the needed size */
    ptl->buf = (double*)realloc(ptl->buf, (size_t)cnt *sizeof(double));
  }
  ptl->msds = (double**)malloc((size_t)(ptl->rowcnt+1)*sizeof(double*));
  if (!ptl->msds) SCN_ERROR(scan, E_NOMEM);
  ptl->msds[0] = m = ptl->buf;  /* create the row vector */
  for (i = 0; ++i <= ptl->rowcnt; )
    ptl->msds[i] = m += ptl->clscnt;
  SCN_CHAR(scan, '}');          /* consume '}' (end of m.s. degrees) */
  SCN_CHAR(scan, ';');          /* consume ';' */
  SCN_CHAR(scan, '}');          /* consume '}' (end of cluster set) */
  SCN_CHAR(scan, ';');          /* consume ';' */
  return 0;                     /* return 'ok' */
}  /* parse() */

/*--------------------------------------------------------------------*/

PTLSET* ptl_parse (SCANNER *scan)
{                               /* --- parse a p.t.-less cluster set */
  PTLSET *ptl;                  /* created prototype-less cluster set */

  assert(scan);                 /* check the function argument */
  scn_setmsgs(scan, msgs, (int)(sizeof(msgs)/sizeof(*msgs)));
  scn_first(scan);              /* get first token */
  ptl = (PTLSET*)malloc(sizeof(PTLSET));
  if (!ptl) { scn_error(scan, E_NOMEM); return NULL; }
  ptl->msds  = ptl->pows  = NULL;
  ptl->buf   = ptl->dists = ptl->wgts = NULL;
  ptl->ids   = NULL;            /* create the base structure */
  if (parse(ptl, scan) != 0) {  /* parse the cluster set */
    ptl_delete(ptl); return NULL; }
  ptl->dists = (double*)malloc((size_t)ptl->rowcnt *sizeof(double));
  ptl->ids   = (DIMID*) malloc((size_t)ptl->rowcnt *sizeof(DIMID));
  if (!ptl->dists || !ptl->ids) {
    ptl_delete(ptl); scn_error(scan, E_NOMEM); return NULL; }
  return ptl;                   /* return the created cluster set */
}  /* ptl_parse() */
